#!/usr/bin/env python3
import argparse
import pickle
import math
import numpy as np
import re
import sys

import sloika.layers

parser = argparse.ArgumentParser()
parser.add_argument('--id', default='' , help='Identifier for model names')
parser.add_argument('model', help='Pickle to read model from')

EMBEDDING_MATRIX = np.array([[1.0, 0.0, -1.0 / np.sqrt(2.0)],
                             [-1.0, 0.0, -1.0 / np.sqrt(2.0)],
                             [0.0, 1.0, 1.0 / np.sqrt(2.0)],
                             [0.0, -1.0, 1.0 / np.sqrt(2.0)]], dtype=np.float32)


trim_trailing_zeros = re.compile('0+p')

def small_hex(f):
    hf = float(f).hex()
    return trim_trailing_zeros.sub('p', hf)


def process_column(v, pad):
    """ process and pad """
    return [small_hex(f) for f in v] + [small_hex(0.0)] * pad


def cformatM(fh, name, X, nr=None, nc=None):
    nrq = int(math.ceil(X.shape[1] / 4.0))
    pad = nrq * 4 - X.shape[1]
    lines = map(lambda v: ', '.join(process_column(v, pad)), X)

    if nr is None:
        nr = X.shape[1]
    else:
        nrq = int(math.ceil(nr / 4.0))
    if nc is None:
        nc = X.shape[0]

    nelt = 4 * nrq * nc

    fh.write('float {}[{}] = {}\n'.format('__' + name, nelt, '{'))
    fh.write('\t' + ',\n\t'.join(lines))
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', nr, nrq, nc, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def cformatV(fh, name, X):
    nrq = int(math.ceil(X.shape[0] / 4.0))
    pad = nrq * 4 - X.shape[0]
    lines = ', '.join(list(map(lambda f: small_hex(f), X)) + [small_hex(0.0)] * pad)
    fh.write('float {}[] = {}\n'.format( '__' + name, '{'))
    fh.write('\t' + lines)
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[0], nrq, 1, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def process_convolution(layer, layerid):
    assert isinstance(layer, sloika.layers.Convolution)

    filterW =  layer.W.get_value()
    nfilter, nfeature , winlen = filterW.shape
    nfeatureCeil4 = 4 * int(math.ceil(nfeature / 4.0))
    cformatM(sys.stdout, '{}W'.format(layerid), filterW.transpose(0, 2, 1).reshape(-1, nfeature),
             nr = (winlen - 1) * nfeatureCeil4 + nfeature, nc=nfilter)
    cformatV(sys.stdout, '{}b'.format(layerid), layer.b.get_value().reshape(-1))
    sys.stdout.write("const int {}stride = {};\n".format(layerid, layer.stride))
    sys.stdout.write("""const size_t _{}nfilter = {};
const size_t _{}winlen = {};
""".format(layerid, nfilter, layerid, winlen))


if __name__ == '__main__':
    args = parser.parse_args()
    modelid = args.id + '_'

    with open(args.model, 'rb') as fh:
        network = pickle.load(fh)
    assert network.major_version >= 2


    sys.stdout.write("""#pragma once
#ifndef NANONET_SQUIGGLE_{}MODEL_H
    #define NANONET_SQUIGGLE_{}MODEL_H
    #include <assert.h>
    #include "../util.h"
""".format(modelid.upper(), modelid.upper()))


    #  Embedding layer
    cformatM(sys.stdout, 'embed_squiggle_{}W'.format(modelid), EMBEDDING_MATRIX)


    """ first convolution
    """
    process_convolution(network.sublayers[0], 'conv1_squiggle_{}'.format(modelid))
    assert isinstance(network.sublayers[1], sloika.layers.Residual)
    process_convolution(network.sublayers[1].sublayers[0], 'conv2_squiggle_{}'.format(modelid))
    assert isinstance(network.sublayers[2], sloika.layers.Residual)
    process_convolution(network.sublayers[2].sublayers[0], 'conv3_squiggle_{}'.format(modelid))
    assert isinstance(network.sublayers[3], sloika.layers.Residual)
    process_convolution(network.sublayers[3].sublayers[0], 'conv4_squiggle_{}'.format(modelid))
    assert isinstance(network.sublayers[4], sloika.layers.Residual)
    process_convolution(network.sublayers[4].sublayers[0], 'conv5_squiggle_{}'.format(modelid))
    process_convolution(network.sublayers[5], 'conv6_squiggle_{}'.format(modelid))

    sys.stdout.write('\n#endif /* NANONET_SQUIGGLE_{}MODEL_H */\n'.format(modelid.upper()))



#!/usr/bin/env python3
import argparse
import numpy as np
import pickle
import sys

from sloika.cmdargs import AutoBool, FileExists, Positive
import sloika.layers as layers
from sloika.helpers import objwalk, set_at_path


parser = argparse.ArgumentParser('Hack temperature into sloika model')
parser.add_argument('--temperature1', default=1.0, type=Positive(float),
                    help='Temperature for softmax weights')
parser.add_argument('--temperature2', default=1.0, type=Positive(float),
                    help='Temperature for softmax bias')
parser.add_argument('model', action=FileExists, help='Pickled sloika model to upgrade')
parser.add_argument('output', help='Output file to write to')


def change_softmax(network, weight_f, bias_f):

    for path, value in objwalk(network, types=(layers.Layer)):
        if isinstance(value, layers.Softmax):
            W = value.W.get_value(borrow=True)
            W *= weight_f
            b = value.b.get_value(borrow=True)
            b *= bias_f
        set_at_path(network, path, value)

    return network


if __name__ == '__main__':
    args = parser.parse_args()

    try:
        with open(args.model, 'rb') as fh:
            network = pickle.load(fh)
        sys.stdout.write('Loaded python 3 pickle from file `{}`\n'.format(args.model))
    except UnicodeDecodeError:
        with open(args.model, 'rb') as fh:
            network = pickle.load(fh, encoding='latin1')
        sys.stdout.write('Loaded python 2 pickle from file `{}`\n'.format(args.model))

    top_level_object_name = type(network).__name__
    assert top_level_object_name == 'Serial', top_level_object_name

    network = change_softmax(network,
                             np.reciprocal(args.temperature1),
                             np.reciprocal(args.temperature2))

    with open(args.output, 'wb') as fo:
        pickle.dump(network, fo)

    sys.stdout.write('Written python 3 pickle to file `{}`\n'.format(args.output))



#!/usr/bin/env python3
import argparse
import pickle
import math
import numpy as np
import re
import sys

parser = argparse.ArgumentParser()
parser.add_argument('model', help='Pickle to read model from')


model_file = sys.argv[1]

trim_trailing_zeros = re.compile('0+p')

def small_hex(f):
    hf = float(f).hex()
    return trim_trailing_zeros.sub('p', hf)


def process_column(v, pad):
    """ process and pad """
    return [small_hex(f) for f in v] + [small_hex(0.0)] * pad


def cformatM(fh, name, X, nr=None, nc=None):
    nrq = int(math.ceil(X.shape[1] / 4.0))
    pad = nrq * 4 - X.shape[1]
    lines = map(lambda v: ', '.join(process_column(v, pad)), X)

    if nr is None:
        nr = X.shape[1]
    else:
        nrq = int(math.ceil(nr / 4.0))
    if nc is None:
        nc = X.shape[0]

    fh.write('float {}[] = {}\n'.format('__' + name, '{'))
    fh.write('\t' + ',\n\t'.join(lines))
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', nr, nrq, nc, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def cformatV(fh, name, X):
    nrq = int(math.ceil(X.shape[0] / 4.0))
    pad = nrq * 4 - X.shape[0]
    lines = ', '.join(list(map(lambda f: small_hex(f), X)) + [small_hex(0.0)] * pad)
    fh.write('float {}[] = {}\n'.format( '__' + name, '{'))
    fh.write('\t' + lines)
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[0], nrq, 1, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


if __name__ == '__main__':
    args = parser.parse_args()
    modelid = 'resgru_'

    with open(args.model, 'rb') as fh:
        network = pickle.load(fh, encoding='latin1')
    network_major_version = network.version[0] if isinstance(network.version, tuple) else network.version
    assert network_major_version >= 1, "Sloika model must be version >= 1.  Perhaps you need to run Sloika's model_upgrade.py"

    sys.stdout.write("""#pragma once
    #ifndef NANONET_RGRGR_{}MODEL_H
    #define NANONET_RGRGR_{}MODEL_H
    #include <assert.h>
    #include "../util.h"
    """.format(modelid.upper(), modelid.upper()))

    """ Convolution layer
    """

    filterW =  network.sublayers[0].W.get_value()
    nfilter, _ , winlen = filterW.shape
    cformatM(sys.stdout, 'conv_rgrgr_{}W'.format(modelid), filterW.reshape(-1, 1), nr = winlen * 4 - 3, nc=nfilter)
    cformatV(sys.stdout, 'conv_rgrgr_{}b'.format(modelid), network.sublayers[0].b.get_value().reshape(-1))
    sys.stdout.write("const int conv_rgrgr_{}stride = {};\n".format(modelid, network.sublayers[0].stride))
    sys.stdout.write("""const size_t _conv_rgrgr_{}nfilter = {};
    const size_t _conv_rgrgr_{}winlen = {};
    """.format(modelid, nfilter, modelid, winlen))

    """  Backward GRU (first layer)
    """
    gru1 = network.sublayers[1].sublayers[0].sublayers[0]
    cformatM(sys.stdout, 'gruB1_rgrgr_{}iW'.format(modelid), gru1.iW.get_value())
    cformatM(sys.stdout, 'gruB1_rgrgr_{}sW'.format(modelid), gru1.sW.get_value())
    cformatM(sys.stdout, 'gruB1_rgrgr_{}sW2'.format(modelid), gru1.sW2.get_value())
    cformatV(sys.stdout, 'gruB1_rgrgr_{}b'.format(modelid), gru1.b.get_value().reshape(-1))

    """  Forward GRU (second layer)
    """
    gru2 = network.sublayers[2].sublayers[0]
    cformatM(sys.stdout, 'gruF2_rgrgr_{}iW'.format(modelid), gru2.iW.get_value())
    cformatM(sys.stdout, 'gruF2_rgrgr_{}sW'.format(modelid), gru2.sW.get_value())
    cformatM(sys.stdout, 'gruF2_rgrgr_{}sW2'.format(modelid), gru2.sW2.get_value())
    cformatV(sys.stdout, 'gruF2_rgrgr_{}b'.format(modelid), gru2.b.get_value().reshape(-1))

    """ backward GRU(third layer)
    """
    gru3 = network.sublayers[3].sublayers[0].sublayers[0]
    cformatM(sys.stdout, 'gruB3_rgrgr_{}iW'.format(modelid), gru3.iW.get_value())
    cformatM(sys.stdout, 'gruB3_rgrgr_{}sW'.format(modelid), gru3.sW.get_value())
    cformatM(sys.stdout, 'gruB3_rgrgr_{}sW2'.format(modelid), gru3.sW2.get_value())
    cformatV(sys.stdout, 'gruB3_rgrgr_{}b'.format(modelid), gru3.b.get_value().reshape(-1))

    """  Forward GRU (fourth layer)
    """
    gru4 = network.sublayers[4].sublayers[0]
    cformatM(sys.stdout, 'gruF4_rgrgr_{}iW'.format(modelid), gru4.iW.get_value())
    cformatM(sys.stdout, 'gruF4_rgrgr_{}sW'.format(modelid), gru4.sW.get_value())
    cformatM(sys.stdout, 'gruF4_rgrgr_{}sW2'.format(modelid), gru4.sW2.get_value())
    cformatV(sys.stdout, 'gruF4_rgrgr_{}b'.format(modelid), gru4.b.get_value().reshape(-1))

    """ backward GRU(fifth layer)
    """
    gru5 = network.sublayers[5].sublayers[0].sublayers[0]
    cformatM(sys.stdout, 'gruB5_rgrgr_{}iW'.format(modelid), gru5.iW.get_value())
    cformatM(sys.stdout, 'gruB5_rgrgr_{}sW'.format(modelid), gru5.sW.get_value())
    cformatM(sys.stdout, 'gruB5_rgrgr_{}sW2'.format(modelid), gru5.sW2.get_value())
    cformatV(sys.stdout, 'gruB5_rgrgr_{}b'.format(modelid), gru5.b.get_value().reshape(-1))
    """ Softmax layer
    """
    nstate = network.sublayers[6].W.get_value().shape[0]
    shuffle = np.append(np.arange(nstate - 1) + 1, 0)
    cformatM(sys.stdout, 'FF_rgrgr_{}W'.format(modelid), network.sublayers[6].W.get_value()[shuffle])
    cformatV(sys.stdout, 'FF_rgrgr_{}b'.format(modelid), network.sublayers[6].b.get_value()[shuffle])

    sys.stdout.write('#endif /* NANONET_RGRGR_{}_MODEL_H */'.format(modelid.upper()))



__version__ = '1.4.0'

import argparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import itertools
import functools
import h5py
import numpy as np
import os

from libscrappy import ffi, lib

ftype = np.float32
size_ftype = np.dtype(ftype).itemsize
vsize = 4  # SSE vector length


def _none_if_null(p):
    # convert cffi NULL to None
    if p == ffi.NULL:
        p = None
    return p


def _gsp():
    # allowing up to 9mer, we can have up to base7 and be unique
    alpha_len = range(4,8)
    kmer_len = range(1,10)
    pairs = [(a, k) for a, k in itertools.product(alpha_len, kmer_len)]
    state_lookup = {a**k: (a, k) for a, k in pairs}
    assert len(state_lookup) == len(pairs) # check all are unique

    def guess_state_properties(nstate):
        """Find likely kmer length and alphabet size from transducer state-space size.

        :param nstate: number of states in transducer.

        :returns: alphabet size, kmer length.
        """
        nkmers = nstate - 1 # for stay
        return state_lookup[nkmers]
    return guess_state_properties

guess_state_properties = _gsp()


class RawTable(object):
    def __init__(self, data, start=0, end=None):
        """Representation of a scrappie `raw_table`.

        :param data: `nd.array` containing raw data.

        ..note:: The class stores a reference to a contiguous numpy array of
            the correct type to be passed to the extension library. The class
            provides safety against the original data being garbage collected.
            To obtain an up-to-date (possibly trimmed and scaled) copy of the
            data use `raw_table.data(as_numpy=True)`.
        """
        if end is None:
            end = len(data)

        self._data = np.ascontiguousarray(data.astype(ftype, order='C', copy=True))
        rt = ffi.new('raw_table *')
        rt.uuid = ffi.NULL
        rt.n = len(self._data)
        rt.start = start
        rt.end = end
        rt.raw = ffi.cast("float *", ffi.from_buffer(self._data))
        self._rt = rt[0]

    def data(self, as_numpy=False):
        """Current data as either C object or realised numpy copy.

        :param as_numpy: If True, return a numpy copy of the current data. If
            currently selected range empty, a (0,) shape array is returned.
        """
        if as_numpy:
            return np.copy(self._data[self.start:self.end])
        else:
            return self._rt

    @property
    def start(self):
        """Currently set start sample."""
        return self._rt.start

    @property
    def end(self):
        """Currently set end sample."""
        return self._rt.end

    def trim(self, start=200, end=10, varseg_chunk=100, varseg_thresh=0.0):
        """Trim data.

        :param start: lower bound on number of samples to trim from start
        :param end: lower bound on number of samples to trim from end
        :param varseg_chunk: chunk size for analysing variance of signal
        :param varseq_thresh: quantile to be calculated to use for thresholding

        """
        self._rt = _trim_raw(
            self._rt, start=start, end=end,
            varseg_chunk=varseg_chunk, varseg_thresh=varseg_thresh
        )
        return self

    def scale(self):
        """Normalize data using med/mad scaling."""
        self._rt = _scale_raw(self._rt)
        return self



def _trim_raw(rt, start=200, end=10, varseg_chunk=100, varseg_thresh=0.0):
    """Trim a `raw_table`.

    :param rt: a `raw_table`.
    :param start: lower bound on number of samples to trim from start
    :param end: lower bound on number of samples to trim from end
    :param varseg_chunk: chunk size for analysing variance of signal
    :param varseg_thresh: quantile to be calculated to use for thresholding

    :returns: new scrappie raw data structure.
    """
    rt = _none_if_null(lib.trim_raw_by_mad(rt, varseg_chunk, varseg_thresh))
    if rt is not None:
        rt.start = rt.start + start if (rt.n - rt.start) > start else rt.n
        rt.end = rt.end - end if (rt.end > end) else 0
    if rt.start >= rt.end:
        rt.start, rt.end = 0, 0
        # n.b. at this point the equivalent trim_and_segment_raw() from
        #      scrappie_common.c frees the data array.
    return rt


def _scale_raw(rt):
    """Scale a `raw_table` in place.

    :param rt: `raw_table` to scale`.

    :returns: input (this is solely to be explicit, `rt` is modified in place).
    """
    lib.medmad_normalise_array(rt.raw + rt.start, rt.end - rt.start)
    return rt


class ScrappyMatrix(object):
    def __init__(self, scrappy_matrix):
        """Container to manage lifetime of a bare scrappie_matrix.
        Can be initialised by a pointer to a `scrappy_matrix` or a numpy `ndarray`.
        """
        if isinstance(scrappy_matrix, np.ndarray):
            self._data = _numpy_to_scrappy_matrix(scrappy_matrix)
        elif isinstance(scrappy_matrix, ffi.CData):
            self._data = scrappy_matrix
        else:
            raise TypeError('ScrappyMatrix can only be contructed from a '
                '`scrappy_matrix pointer or `ndarray`.')

    def __del__(self):
        _free_matrix(self._data)

    @property
    def shape(self):
        """Tuple (columns, rows)

        .. note:: The rows value is of the realised array. The actual data is
            stored padded for ease of use with SSE vectors.
        """
        return self._data.nc, self._data.nr

    def data(self, as_numpy=False, sloika=True):
        """Current data as either C object or realised numpy copy. In the
        latter case, padding due to SSE vector use is removed.

        :param as_numpy: If True, return a numpy copy of the current data.
        :param sloika: return sloika compatible posterior matrix (stay is
            first state). Only valid when `as_numpy` is `True`.

        :returns: a contiguous `np.ndarray` of shape (blocks, states).

        """
        if as_numpy:
            return _scrappie_to_numpy(self._data, sloika=sloika)
        else:
            return self._data

    def __getitem__(self, slice):
        return ScrappyMatrixView(self, slice)

    def __len__(self):
        return self.shape[0]


class ScrappyMatrixView(ScrappyMatrix):
    def __init__(self, scrappy_matrix_obj, slice):
        """Container to provide view of `ScrappyMatrix`.
        Won't free underlying data upon garbage collection.

        :param scrappy_matrix_obj: `ScrappyMatrix` instance.
        :param slice: slice
        """
        n_elems = scrappy_matrix_obj.shape[0]
        start = slice.start if slice.start is not None else 0
        stop = slice.stop if slice.stop is not None else n_elems

        if slice.step is not None and slice.step != 1:
            raise ValueError('stride must be 1 for a view')

        if start > stop:
            raise IndexError('start should be smaller than stop')

        for i in start, stop:
            if i > n_elems:
                raise IndexError('index {} is out of bounds for axis 0 with size {}'.format(i, n_elems))

        attrs = ['nr', 'nrq', 'nc', 'stride', 'data']
        init_data = [getattr(scrappy_matrix_obj._data, a) for a in attrs]
        self._data = ffi.new("scrappie_matrix", init=init_data)
        self._data.data.f += start * self._data.stride
        self._data.nc = stop - start

    def __del__(self):
        pass  # This is a view, we don't want underlying data garbage collected


def _numpy_to_scrappy_matrix(numpy_array):
    """Convert a `ndarray` to a bare `scrappie_matrix`"""
    nc = numpy_array.shape[0]
    nr = numpy_array.shape[1]

    data = np.ascontiguousarray(numpy_array.astype(ftype, order='C', copy=False))
    buf = ffi.cast("float *", data.ctypes.data)
    return lib.mat_from_array(buf, nr, nc)


def _free_matrix(matrix):
    """Free a `scrappie_matrix`.

    :param matrix: a scrappie matrix

    :returns: `None`
    """
    lib.free_scrappie_matrix(matrix)


def _scrappie_to_numpy(matrix, sloika=True):
    """Convert a `scrappie_matrix` to a numpy array. Removes padding due to
    SSE vectors and optionally reorders states.

    :param matrix: a `scrappie_matrix` to use as source data.
    :param sloika: return sloika compatible matrix (stay is first state).

    :returns: a contiguous `np.ndarray` of shape (blocks, states).

    ..note:: a copy of the data is made, so the input matrix should be freed
        at the earliest convenience with `_free_matrix`.
    """
    np_matrix = np.frombuffer(ffi.buffer(
        matrix.data.f, size_ftype * vsize * matrix.nrq * matrix.nc),
        dtype=ftype
    ).reshape(matrix.nc, vsize * matrix.nrq)

    nblock, nstate = matrix.nc, matrix.nr
    np_matrix = np_matrix[:, :nstate]

    # sloika requires stay state first
    if sloika:
        p1 = np_matrix[:, nstate - 1:nstate]
        p2 = np_matrix[:, 0:nstate - 1]
        np_matrix = np.hstack((p1, p2))
    np_matrix = np.ascontiguousarray(np_matrix)
    return np_matrix


def calc_post(rt, model='rgrgr_r94', min_prob=1e-6, log=True, tempW=1.0, tempb=1.0):
    """Run a network network to obtain class probabilities.

    :param rt: a `RawTable`.
    :param min_prob: minimum bound of probabilites.
    :param log: return log-probabilities.

    :returns: a `ScrappyMatrix`
    """
    if not log and model == 'rnnrf_r94':
        raise ValueError("Returning non-log transformed matrix not supported for model type 'rnnrf_r94'.")
    if not isinstance(rt, RawTable):
        raise TypeError('`rt` should be a RawTable.')

    try:
        network = _models_[model]
    except KeyError:
        raise KeyError("Model type '{}' not recognised.".format(model))
    else:
        matrix = _none_if_null(network(rt.data(), min_prob, tempW, tempb, log))
        if matrix is None:
            raise RuntimeError('An unknown error occurred during posterior calculation.')
        else:
            return ScrappyMatrix(matrix)


def decode_post(post, model='rgrgr_r94', **kwargs):
    """Decode a posterior to retrieve basecall, score, and states. This
    function merely dispatches to a relevant function governed by the model.

    :param post: a `ScrappyMatrix` containing matrix to be decoded.
    :param model: model type.
    :param **kwargs: See the functions `_decode_...` for kwargs
        relevant to each model.

    :returns: tuple containing (call, score, call positions per raw block).
    """
    if not isinstance(post, ScrappyMatrix):
        raise TypeError('`post` should be a ScrappyMatrix.')
    try:
        decoder = _decoders_[model]
    except KeyError:
        raise KeyError("Model type '{}' not recognised.".format(model))
    else:
        return decoder(post, **kwargs)


def _decode_post(post, stay_pen=0.0, skip_pen=0.0, local_pen=2.0, use_slip=False):
    """Decode a posterior using Viterbi algorithm for transducer.

    :param post: a `ScrappyMatrix` containing transducer posteriors.
    :param stay_pen: penalty for staying.
    :param skip_pen: penalty for skipping a base.
    :param local_pen: penalty for local basecalling.
    :param use_slip: allow slipping (movement more than 2 bases).

    :returns: tuple containing (call, score, call positions per raw block).
    """
    nblock, nstate = post.shape

    path = ffi.new("int[{}]".format(nblock + 1))
    score = lib.decode_transducer(
        post.data(), stay_pen, skip_pen, local_pen,
        path, use_slip
    )

    pos = np.zeros(nblock + 1, dtype=np.int32)
    p_pos = ffi.cast("int *", pos.ctypes.data)
    basecall = lib.overlapper(path, nblock + 1, nstate - 1, p_pos)

    return ffi.string(basecall).decode(), score, pos


def _decode_post_crf(post):
    """Decode a posterior using Viterbi algorithm for conditional random field.

    :param post: a `ScrappyMatrix` containing CRF transitions.

    :returns: tuple containing (basecall, score, call positions per raw data block).
    """
    nblock, nstate = post.shape

    path = ffi.new("int[{}]".format(nblock + 1))
    score = lib.decode_crf(post.data(), path)

    pos = np.ascontiguousarray(np.zeros(nblock + 1, dtype=np.int32))
    p_pos = ffi.cast("int *", ffi.from_buffer(pos))
    basecall = lib.crfpath_to_basecall(path, nblock, p_pos)

    return ffi.string(basecall).decode(), score, pos


# Network and decoder functions used above
_models_ = {
    'rgrgr_r94': lib.nanonet_rgrgr_r94_posterior,
    'rgrgr_r941': lib.nanonet_rgrgr_r941_posterior,
    'rgrgr_r10': lib.nanonet_rgrgr_r10_posterior,
    'rnnrf_r94': lib.nanonet_rnnrf_r94_transitions,
}

_squiggle_models_ = {
    'squiggle_r94': lib.squiggle_r94,
    'squiggle_r94_rna': lib.squiggle_r94_rna,
    'squiggle_r10': lib.squiggle_r10,
}

_decoders_ = {
    'rgrgr_r94': _decode_post,
    'rgrgr_r941': _decode_post,
    'rgrgr_r10': _decode_post,
    'rnnrf_r94': _decode_post_crf,
}


def get_model_stride(model):
    """Obtain the stride length of a model from its name.

    :param model: model name:

    :returns: the model stride.
    """
    stride = lib.get_raw_model_stride_from_string(model.encode())
    if stride == -1:
        raise ValueError("Invalid scrappie model '{}'.")
    return stride


def basecall_raw(data, model='rgrgr_r94', with_base_probs=False, **kwargs):
    """Basecall from raw data in a numpy array to demonstrate API.

    :param data: `ndarray` containing raw signal data.
    :param model: model to use in calculating basecall.
    :param with_base_probs: calculate per-block base (ACGT-) probabilities.
    :param kwargs: kwargs passed to `decode_post`.

    :returns: tuple containing: (basecall, score, per-block call positions
        data start index, data end index, base probs). The last item will
        be `None` for `with_base_probs == False`.
    """
    if with_base_probs and model != 'rnnrf_r94':
        ValueError("Base probabilities can only be returned for model 'rnnrf_r94'.")

    raw = RawTable(data)
    raw.trim().scale()

    post = calc_post(raw, model, log=True)
    seq, score, pos = decode_post(post, model, **kwargs)

    base_probs = None
    if with_base_probs:
        base_post = lib.posterior_crf(post.data())
        base_probs = _scrappie_to_numpy(base_post, sloika=False)
        _free_matrix(base_post)

    return seq, score, pos, raw.start, raw.end, base_probs


def sequence_to_squiggle(sequence, model='squiggle_r94', rescale=False):
    """Simulate a squiggle from a base sequence.

    :param sequence: base sequence to model.
    :param model: model to use in simulating squiggle.
    :param rescale:
    :param as_numpy: return a numpy array rather than a `scrappie_matrix`.

    :returns: a simulated squiggle as either a `ScrappyMatrix` or an ndarray.

    """

    seq_len = len(sequence)
    seq = _none_if_null(lib.encode_bases_to_integers(sequence.encode(), seq_len, 1))
    if seq is None:
        raise RuntimeError('An unknown error occurred whilst encoding sequence.')

    try:
        squiggle_model = _squiggle_models_[model]
    except KeyError:
        raise KeyError("Squiggle model type '{}' not recognised.".format(model))
    else:
        squiggle = _none_if_null(squiggle_model(seq, seq_len, rescale))
    if squiggle is None:
        raise RuntimeError('An unknown error occurred whilst generating squiggle.')

    return ScrappyMatrix(squiggle)


def map_signal_to_squiggle(data, sequence, model='squiggle_r94', rate=1.0,
                           back_prob=0.0, local_pen=2.0, skip_pen=5000.0,
                           min_score=5.0):
    """Align a squiggle to a sequence using a simulated squiggle.

    :param data: `ndarray` containing raw signal data.
    :param sequence: base sequence to which to align data.
    :param model: model to use in simulating squiggle.
    :param rate: rate of translocation relative to squiggle model
    :param back_prob: probability of backward movement.
    :param local_pen: penalty for local alignment.
    :param skip_pen: penalty for skipping position in sequence.
    :param min_score: floor on match score.

    :returns: tuple containing (alignment score, alignment path)
    """
    raw = RawTable(data)
    raw.trim().scale()

    squiggle = sequence_to_squiggle(sequence, model=model)

    path = np.ascontiguousarray(np.zeros(raw._rt.n, dtype=np.int32))
    p_path = ffi.cast("int32_t *", ffi.from_buffer(path))

    score = lib.squiggle_match_viterbi(raw.data(), rate, squiggle.data(), back_prob,
                                       local_pen, skip_pen, min_score, p_path)

    return score, path


def map_post_to_sequence(post, sequence, stay_pen=0, skip_pen=0, local_pen=4.0,
                         viterbi=False, path=False, bands=None):
    """Block-based local-global alignment of a squiggle to a sequence using
    either Forward or Viterbi algorithm. For the latter the Viterbi path can
    optionally be calculated.

    :param post: a `ScrappyMatrix` containing log-probabilities (as from
       `calc_post`).
    :param sequence: a base sequence which to map.
    :param stay_pen: penalty for zero-state movement from one block to next.
    :param skip_pen: penalty for two-state movement from one block to next.
    :param local_pen: penalty for local alignment through blocks
    :param viterbi: use Viterbi algorithm rather than forward.
    :param path: calculate alignment path (only valid for `viterbi==True`
        and `bands==None`).
    :param bands: two sequences containing lower and upper extremal allowed
        positions for each block. Should be length corresponding to number
        of blocks of `post`. If a single number is given, a diagonal band with
        width 2 * `bands` * #states / #blocks will be used. If `None` is given
        banding is not used (a full DP matrix is evaluated).

    :returns: (score, path), (or (None, *) in the case of failure).

    ..note:: if `viterbi`==False or `path`==False, the returned path will
        be `None`.
    """

    if path and not viterbi:
        raise ValueError('Cannot calulate path with `viterbi==False`.')
    if not isinstance(post, ScrappyMatrix):
        raise TypeError('`post` should be a ScrappyMatrix.')

    nblock, nstate = post.shape
    alpha_len, kmer_len = guess_state_properties(nstate)

    seq_len = len(sequence) - kmer_len + 1
    p_seq = _none_if_null(lib.encode_bases_to_integers(sequence.encode(), len(sequence), kmer_len))
    if p_seq is None:
        raise RuntimeError('An unknown error occurred whilst encoding sequence.')

    if viterbi and path:
        path_data = np.zeros(nblock, dtype=np.int32)
        p_path = ffi.cast("int *", ffi.from_buffer(path_data))
    else:
        path_data = None
        p_path = ffi.NULL

    if bands is None:
        if viterbi:
            score = lib.map_to_sequence_viterbi(
                post.data(), stay_pen, skip_pen, local_pen, p_seq, seq_len, p_path)
        else:
            score = lib.map_to_sequence_forward(
                post.data(), stay_pen, skip_pen, local_pen, p_seq, seq_len)
    else:
        if isinstance(bands, int):
            # create a monotonic diagonal band
            gradient = seq_len / nblock
            bands = 2 * bands * gradient
            hband = bands / 2
            bands = [np.ascontiguousarray(np.array(x, dtype=np.uintp)) for x in (
                [(max(0,       x * gradient - hband)) for x in range(nblock)],
                [(min(seq_len, x * gradient + hband)) for x in range(nblock)]
            )]
        elif len(bands) == 2:
            bands = [
                np.ascontiguousarray(x, dtype=np.uintp) for x in bands]
        else:
            raise ValueError('`bands` should be `None`, an integer, or length 2.')

        p_poslow, p_poshigh = (
            ffi.cast("size_t *", ffi.from_buffer(x)) for x in bands)
        if not lib.are_bounds_sane(p_poslow, p_poshigh, nblock, seq_len):
            raise ValueError('Supplied banding structure is not valid.')

        if viterbi:
            func = lib.map_to_sequence_viterbi_banded
        else:
            func = lib.map_to_sequence_forward_banded
        score = func(
            post.data(), stay_pen, skip_pen, local_pen, p_seq, seq_len, p_poslow, p_poshigh)

    score = _none_if_null(score)
    if score is None:
        raise RuntimeError('An unknown error occurred during alignment.')

    return score, path_data


def _raw_gen(filelist):
    for fname in filelist:
        with h5py.File(fname, 'r') as h:
            pass
        try:
            data = None
            with h5py.File(fname, 'r') as h:
                base = 'Raw/Reads'
                read_name = list(h[base].keys())[0]
                data = h['{}/{}/Signal'.format(base, read_name)][()].astype(np.float32)
                meta = h['/UniqueGlobalKey/channel_id'].attrs
                raw_unit = meta['range'] / meta['digitisation']
                data = (data + meta['offset']) * raw_unit
        except:
            raise RuntimeError('Failed to read signal data from {}.'.format(fname))
        else:
            yield os.path.basename(fname), data


def _basecall():
    # Entry point for testing/demonstration.
    parser = argparse.ArgumentParser(
        description="Basecall a single .fast5.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('fast5', nargs='+',
        help='path to .fast5s to basecall.')
    parser.add_argument('model', choices=list(_models_.keys()),
        help='Choice of model.')
    parser.add_argument('--threads', default=None, type=int,
        help='Number of threads to use.')
    parser.add_argument('--process', action='store_true',
        help='Use ProcesPool rather than ThreadPool.')

    args = parser.parse_args()

    worker = functools.partial(basecall_raw, model=args.model)
    if args.threads is None:
        for fname, data in _raw_gen(args.fast5):
            seq, score, _, start, end, _ = worker(data)
            print(">{} {} {}-{}\n{}".format(fname, score, start, end, seq))
    else:
        iter0, iter1 = itertools.tee(_raw_gen(args.fast5))
        Executor = ProcessPoolExecutor if args.process else ThreadPoolExecutor
        with Executor(max_workers=args.threads) as executor:
            datas = (x[1] for x in iter0)
            fnames = (x[0] for x in iter1)
            results = executor.map(worker, datas)
            for fname, (seq, score, _, start, end, _) in zip(fnames, results):
                print(">{} {} {}-{}\n{}".format(fname, score, start, end, seq))




#!/usr/bin/env python3
import pickle
import math
import numpy as np
import re
import sys


model_file = sys.argv[1]
trim_trailing_zeros = re.compile('0+p')

def small_hex(f):
    hf = float(f).hex()
    return trim_trailing_zeros.sub('p', hf)

def process_column(v, pad):
    """ process and pad """
    return [small_hex(f) for f in v] + [small_hex(0.0)] * pad


def cformatM(fh, name, X, nr=None, nc=None):
    nrq = int(math.ceil(X.shape[1] / 4.0))
    pad = nrq * 4 - X.shape[1]
    lines = map(lambda v: ', '.join(process_column(v, pad)), X)

    if nr is None:
        nr = X.shape[1]
    else:
        nrq = int(math.ceil(nr / 4.0))
    if nc is None:
        nc = X.shape[0]

    fh.write('float {}[] = {}\n'.format('__' + name, '{'))
    fh.write('\t' + ',\n\t'.join(lines))
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', nr, nrq, nc, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def cformatV(fh, name, X):
    nrq = int(math.ceil(X.shape[0] / 4.0))
    pad = nrq * 4 - X.shape[0]
    lines = ', '.join(list(map(lambda f: small_hex(f), X)) + [small_hex(0.)] * pad)
    fh.write('float {}[] = {}\n'.format( '__' + name, '{'))
    fh.write('\t' + lines)
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[0], nrq, 1, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))



with open(model_file, 'rb') as fh:
    network = pickle.load(fh, encoding='latin1')
assert network.version == 1, "Sloika model must be version 1.  Perhaps you need to run Sloika's model_upgrade.py"

sys.stdout.write("""#pragma once
#ifndef NANONET_RAW_MODEL_H
#define NANONET_RAW_MODEL_H
#include <assert.h>
#include "../util.h"
""")

""" First LSTM layer
"""

filterW =  network.sublayers[0].W.get_value()
nfilter, _ , winlen = filterW.shape
cformatM(sys.stdout, 'conv_raw_W', filterW.reshape(-1, 1), nr = winlen * 4 - 3, nc=nfilter)
cformatV(sys.stdout, 'conv_raw_b', network.sublayers[0].b.get_value().reshape(-1))
sys.stdout.write("const int conv_raw_stride = {};\n".format(network.sublayers[0].stride))
sys.stdout.write("""const size_t _conv_nfilter = {};
const size_t _conv_winlen = {};
""".format(nfilter, winlen))

bigru1 = network.sublayers[1]
gru = bigru1.sublayers[0]
cformatM(sys.stdout, 'gruF1_raw_iW', gru.iW.get_value())
cformatM(sys.stdout, 'gruF1_raw_sW', gru.sW.get_value())
cformatM(sys.stdout, 'gruF1_raw_sW2', gru.sW2.get_value())
cformatV(sys.stdout, 'gruF1_raw_b', gru.b.get_value().reshape(-1))

gru = bigru1.sublayers[1].sublayers[0]
cformatM(sys.stdout, 'gruB1_raw_iW', gru.iW.get_value())
cformatM(sys.stdout, 'gruB1_raw_sW', gru.sW.get_value())
cformatM(sys.stdout, 'gruB1_raw_sW2', gru.sW2.get_value())
cformatV(sys.stdout, 'gruB1_raw_b', gru.b.get_value().reshape(-1))


""" First feed forward layer
"""
assert(network.sublayers[2].insize % 2 == 0)
size = network.sublayers[2].insize // 2
cformatM(sys.stdout, 'FF1_raw_Wf', network.sublayers[2].W.get_value()[:, : size])
cformatM(sys.stdout, 'FF1_raw_Wb', network.sublayers[2].W.get_value()[:, size : 2 * size])
cformatV(sys.stdout, 'FF1_raw_b', network.sublayers[2].b.get_value())


""" Second GRU layer
"""
bigru1 = network.sublayers[3]
gru = bigru1.sublayers[0]
cformatM(sys.stdout, 'gruF2_raw_iW', gru.iW.get_value())
cformatM(sys.stdout, 'gruF2_raw_sW', gru.sW.get_value())
cformatM(sys.stdout, 'gruF2_raw_sW2', gru.sW2.get_value())
cformatV(sys.stdout, 'gruF2_raw_b', gru.b.get_value().reshape(-1))

gru = bigru1.sublayers[1].sublayers[0]
cformatM(sys.stdout, 'gruB2_raw_iW', gru.iW.get_value())
cformatM(sys.stdout, 'gruB2_raw_sW', gru.sW.get_value())
cformatM(sys.stdout, 'gruB2_raw_sW2', gru.sW2.get_value())
cformatV(sys.stdout, 'gruB2_raw_b', gru.b.get_value().reshape(-1))


""" Second feed forward layer
"""
size = network.sublayers[4].insize // 2
assert(network.sublayers[4].insize % 2 == 0)
cformatM(sys.stdout, 'FF2_raw_Wf', network.sublayers[4].W.get_value()[:, : size])
cformatM(sys.stdout, 'FF2_raw_Wb', network.sublayers[4].W.get_value()[:, size : 2 * size])
cformatV(sys.stdout, 'FF2_raw_b', network.sublayers[4].b.get_value())


""" Softmax layer
"""
nstate = network.sublayers[5].W.get_value().shape[0]
shuffle = np.append(np.arange(nstate - 1) + 1, 0)
cformatM(sys.stdout, 'FF3_raw_W', network.sublayers[5].W.get_value()[shuffle])
cformatV(sys.stdout, 'FF3_raw_b', network.sublayers[5].b.get_value()[shuffle])

sys.stdout.write('#endif /* NANONET_RAW_MODEL_H */')



#!/usr/bin/env python3
import pickle
import math
import numpy as np
import re
import sys


model_file = sys.argv[1]
trim_trailing_zeros = re.compile('0+p')


def small_hex(f):
    hf = float(f).hex()
    return trim_trailing_zeros.sub('p', hf)


def process_column(v, pad):
    """ process and pad """
    return [small_hex(f) for f in v] + [small_hex(0.0)] * pad


def cformatM(fh, name, X):
    nrq = int(math.ceil(X.shape[1] / 4.0))
    pad = nrq * 4 - X.shape[1]
    lines = map(lambda v: ', '.join(process_column(v, pad)), X)

    fh.write('float {}[] = {}\n'.format('__' + name, '{'))
    fh.write('\t' + ',\n\t'.join(lines))
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[1], nrq, X.shape[0], nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def cformatV(fh, name, X):
    nrq = int(math.ceil(X.shape[0] / 4.0))
    pad = nrq * 4 - X.shape[0]
    lines = ', '.join(list(map(lambda f: small_hex(f), X)) + [small_hex(0.0)] * pad)
    fh.write('float {}[] = {}\n'.format('__' + name, '{'))
    fh.write('\t' + lines)
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[0], nrq, 1, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def reshape_lstmM(mat):
    _, isize = mat.shape
    return mat.reshape((-1, 4, isize)).transpose([1, 0, 2]).reshape((-1, isize))


def reshape_lstmV(mat):
    return mat.reshape((-1, 4)).transpose().reshape(-1)


with open(model_file, 'rb') as fh:
    network = pickle.load(fh, encoding='latin1')
assert network.version == 1, "Sloika model must be version 1.  Perhaps you need to run Sloika's model_upgrade.py"

sys.stdout.write("""#pragma once
#ifndef NANONET_EVENTS_MODEL_H
#define NANONET_EVENTS_MODEL_H
#include "../util.h"
""")

""" First LSTM layer
"""

bilstm1 = network.sublayers[1]
lstm = bilstm1.sublayers[0]
cformatM(sys.stdout, 'lstmF1_iW', reshape_lstmM(lstm.iW.get_value()))
cformatM(sys.stdout, 'lstmF1_sW', reshape_lstmM(lstm.sW.get_value()))
cformatV(sys.stdout, 'lstmF1_b', reshape_lstmV(lstm.b.get_value().reshape(-1)))
cformatV(sys.stdout, 'lstmF1_p', lstm.p.get_value().reshape(-1))

lstm = bilstm1.sublayers[1].sublayers[0]
cformatM(sys.stdout, 'lstmB1_iW', reshape_lstmM(lstm.iW.get_value()))
cformatM(sys.stdout, 'lstmB1_sW', reshape_lstmM(lstm.sW.get_value()))
cformatV(sys.stdout, 'lstmB1_b', reshape_lstmV(lstm.b.get_value().reshape(-1)))
cformatV(sys.stdout, 'lstmB1_p', lstm.p.get_value().reshape(-1))


""" First feed forward layer
"""
size = network.sublayers[2].insize // 2
cformatM(sys.stdout, 'FF1_Wf', network.sublayers[2].W.get_value()[:, : size])
cformatM(sys.stdout, 'FF1_Wb', network.sublayers[2].W.get_value()[:, size : 2 * size])
cformatV(sys.stdout, 'FF1_b', network.sublayers[2].b.get_value())


""" Second LSTM layer
"""
bilstm2 = network.sublayers[3]
lstm = bilstm2.sublayers[0]
cformatM(sys.stdout, 'lstmF2_iW', reshape_lstmM(lstm.iW.get_value()))
cformatM(sys.stdout, 'lstmF2_sW', reshape_lstmM(lstm.sW.get_value()))
cformatV(sys.stdout, 'lstmF2_b', reshape_lstmV(lstm.b.get_value().reshape(-1)))
cformatV(sys.stdout, 'lstmF2_p', lstm.p.get_value().reshape(-1))

lstm = bilstm2.sublayers[1].sublayers[0]
cformatM(sys.stdout, 'lstmB2_iW', reshape_lstmM(lstm.iW.get_value()))
cformatM(sys.stdout, 'lstmB2_sW', reshape_lstmM(lstm.sW.get_value()))
cformatV(sys.stdout, 'lstmB2_b', reshape_lstmV(lstm.b.get_value().reshape(-1)))
cformatV(sys.stdout, 'lstmB2_p', lstm.p.get_value().reshape(-1))


""" Second feed forward layer
"""
size = network.sublayers[4].insize // 2
cformatM(sys.stdout, 'FF2_Wf', network.sublayers[4].W.get_value()[:, : size])
cformatM(sys.stdout, 'FF2_Wb', network.sublayers[4].W.get_value()[:, size : 2 * size])
cformatV(sys.stdout, 'FF2_b', network.sublayers[4].b.get_value())


""" Softmax layer
"""
nstate = network.sublayers[5].W.get_value().shape[0]
shuffle = np.append(np.arange(nstate - 1) + 1, 0)
cformatM(sys.stdout, 'FF3_W', network.sublayers[5].W.get_value()[shuffle])
cformatV(sys.stdout, 'FF3_b', network.sublayers[5].b.get_value()[shuffle])

sys.stdout.write('#endif /* NANONET_EVENTS_MODEL_H */')



import json
from sys import stdin


first_entry = json.loads(stdin.readline())
keys = list(first_entry.keys())
print('\t'.join(keys))
print('\t'.join( str(v) for v in first_entry.values()))

for line in  stdin.readlines():
    entry = json.loads(line)
    print('\t'.join( str(entry[k]) for k in keys ))



import os

from cffi import FFI

if 'MANYLINUX' in os.environ:
    src_dir = os.path.join('/io', 'src')
    # build-wheels.sh determines these
    libraries=['openblas']
    library_dirs=['/usr/local/lib/']
else:
    if os.path.isfile(os.path.join('..', 'src','decode.h')):
        # assume the git repo
        src_dir = os.path.join('..', 'src')
    elif os.path.isfile(os.path.join('src','decode.h')):
        # else we're from an sdist
        src_dir = 'src'
    else:
        raise IOError('Cannot find scrappie C sources.')
    # this might want to be cblas on some systems
    libraries=['blas']
    library_dirs=[]

ffibuilder = FFI()
ffibuilder.set_source("libscrappy",
    r"""
      #include "decode.h"
      #include "networks.h"
      #include "scrappie_common.h"
      #include "util.h"
      #include "scrappie_seq_helpers.h"
      #include "scrappie_matrix.h"

      int get_raw_model_stride_from_string(const char * modelstr){
        // Obtain the model stride from its str name
        // avoid the intermediate errx from C signalling bad model
        // name with a return of -1.
        const enum raw_model_type modeltype = get_raw_model(modelstr);
        if(modeltype == SCRAPPIE_MODEL_INVALID){
          return -1;
        } else {
          return get_raw_model_stride(modeltype);
        }
      }

    """,
    libraries=libraries,
    library_dirs=library_dirs,
    include_dirs=[src_dir],
    sources=[
        os.path.join(src_dir, '{}.c'.format(x)) for x in
        r'''decode
            event_detection
            layers
            networks
            nnfeatures
            scrappie_common
            scrappie_matrix
            scrappie_seq_helpers
            util'''.split()
    ],
    extra_compile_args=['-std=c99', '-msse3', '-O3']
)

with open('pyscrap.h', 'r') as fh:
    pyscrap_function_prototypes = fh.read()

ffibuilder.cdef("""
typedef struct {
  char * uuid;
  size_t n;
  size_t start;
  size_t end;
  float *raw;
} raw_table;

typedef struct {
  size_t nr, nrq, nc, stride;
  union {
    //__m128 *v; // we don't need this
    float *f;
  } data;
} _Mat;
typedef _Mat *scrappie_matrix;
typedef _Mat const *const_scrappie_matrix;
""" + pyscrap_function_prototypes)

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



import argparse
import numpy as np

parser = argparse.ArgumentParser(description='')
parser.add_argument('file')

def read_crp(filename):
    with open(filename, 'r') as fh:
        nr, nc = [int(x) for x in fh.readline().rstrip().split()]

        mat = np.zeros((nc, nr))
        for col in range(nc):
            mat[col] = [float.fromhex(x) for x in fh.readline().rstrip().split()]

    return mat

if __name__ == '__main__':
    args = parser.parse_args()

    mat = read_crp(args.file)
    print(mat)



import cPickle
import numpy as np
from sloika import features
from untangled import fast5

modelfile = 'model_final.pkl'
fn = 'reads/MINICOL228_20161012_FNFAB42578_MN17976_mux_scan_HG_52221_ch271_read66_strand.fast5'

with open(modelfile, 'r') as fh:
	network = cPickle.load(fh)

iWf = network.layers[1].layers[0].iW.get_value()
sWf = network.layers[1].layers[0].sW.get_value()
sW2f = network.layers[1].layers[0].sW2.get_value()
bf = network.layers[1].layers[0].b.get_value()
iWb = network.layers[1].layers[1].layer.iW.get_value()
sWb = network.layers[1].layers[1].layer.sW.get_value()
sW2b = network.layers[1].layers[1].layer.sW2.get_value()
bb = network.layers[1].layers[1].layer.b.get_value()

with fast5.Reader(fn) as f5:
    ev = f5.get_section_events("template", analysis='Segment_Linear')
    sn = f5.filename_short

ev = ev[50:]
print "Input"
print ev[:10]

features = features.from_events(ev, tag='')
print "Features"
print features[:10]

#  Windowing
w = 3
inMat = features
zeros = np.zeros((w // 2, features.shape[1]))
padMat = np.concatenate([zeros, inMat, zeros], axis=0)
tmp = np.concatenate([padMat[i : 1 + i - w] for i in xrange(w - 1)], axis=1)
feature3 = np.concatenate([tmp, padMat[w - 1 :]], axis=1)
print "Feature 3"
print feature3[:10]

#  GRU forward
def gatefun(x):
	return 1.0 / (1.0 + np.exp(-x))
size = sWf.shape[1]
state = np.zeros(size)
outF = np.zeros((len(feature3), size))
for i, f in enumerate(feature3):
        vI = np.tensordot(f, iWf, axes=(0,1)) + bf
        vS = np.tensordot(state, sWf, axes=(0,1))
        vT = vI[:2 * size] + vS
        vT = vT.reshape((2, -1))

        z = gatefun(vT[0])
        r = gatefun(vT[1])
        y = np.tensordot(r * state, sW2f, axes=(0,1))
        hbar = np.tanh(vI[2 * size:] + y)
        state = z * state + (1 - z) * hbar
	outF[i] = state
print "GRU forward"
print outF[:10,:8]

state = np.zeros(size)
outB = np.zeros((len(feature3), size))
for i, f in enumerate(feature3[::-1]):
        vI = np.tensordot(f, iWb, axes=(0,1)) + bb
        vS = np.tensordot(state, sWb, axes=(0,1))
        vT = vI[:2 * size] + vS
        vT = vT.reshape((2, -1))

        z = gatefun(vT[0])
        r = gatefun(vT[1])
        y = np.tensordot(r * state, sW2b, axes=(0,1))
        hbar = np.tanh(vI[2 * size:] + y)
        state = z * state + (1 - z) * hbar
        outB[i] = state
outB = outB[::-1]
print "GRU backward"
print outB[:10,:8]

inFF = np.concatenate((outF, outB), axis=1)

W1 = network.layers[2].W.get_value()
b1 = network.layers[2].b.get_value()
outFF1 = np.tanh(np.tensordot(inFF, W1, axes=(1,1)) + b1)
print "FF1"
print outFF1[:10,:8]



import os
import sys
import re
from glob import glob
from setuptools import setup, find_packages, Extension


__pkg_name__ = 'scrappy'
__author__ = 'cwright'
__description__ = 'Python bindings to scrappie basecaller'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'


__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_name__))

# Get the version number from __init__.py
verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))


# Get requirements from file, we prefer to have
#   preinstalled these with pip to make use of wheels.
dir_path = os.path.dirname(__file__)
install_requires = []
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    reqs = (
        r.split('#')[0].strip()
        for r in fh.read().splitlines() if not r.startswith('#')
    )
    for req in reqs:
        if req == '':
            continue
        if req.startswith('git+https'):
            req = req.split('/')[-1].split('@')[0]
        install_requires.append(req)

extra_requires = {}
additional_tests_requires = ['nose>=1.0']
extensions = []

classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    'Natural Language :: English',
    'Operating System :: Unix',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Mathematics'
]


if 'MANYLINUX' in os.environ:
    # when building wheel we need to include this
    data_files = [('', ['OPENBLAS_LICENSE'])]
else:
    data_files = []

setup(
    name='scrappie', # scrappy is taken on pypi
    version=__version__,
    url='https://github.com/nanoporetech/scrappie',
    author=__author__,
    author_email='{}@nanoporetech.com'.format(__author__),
    classifiers=classifiers,
    description=__description__,
    long_description=__long_description__,
    long_description_content_type=__long_description_content_type__,
    python_requires='~=3.4',
    license='MPL 2.0',

    dependency_links=[],
    ext_modules=extensions,
    install_requires=install_requires,
    tests_require=install_requires + additional_tests_requires,
    extras_require=extra_requires,
    setup_requires=["cffi>=1.0.0"],
    cffi_modules=["build.py:ffibuilder"],
    # don't include any testing subpackages in dist
    packages=find_packages(exclude=['*.test', '*.test.*', 'test.*', 'test']),
    package_data={},
    zip_safe=False,
    data_files=data_files,
    entry_points={
        'console_scripts': [
            '{} = {}:_basecall'.format(__pkg_name__, __pkg_name__)
        ]
    },
    scripts=[
    ],
    test_suite='nose.collector',
)



#!/usr/bin/env python3
import pickle
import math
import numpy as np
import re
import sys


model_file = sys.argv[1]

trim_trailing_zeros = re.compile('0+p')

def small_hex(f):
    hf = float(f).hex()
    return trim_trailing_zeros.sub('p', hf)


def process_column(v, pad):
    """ process and pad """
    return [small_hex(f) for f in v] + [small_hex(0.0)] * pad


def cformatM(fh, name, X, nr=None, nc=None):
    nrq = int(math.ceil(X.shape[1] / 4.0))
    pad = nrq * 4 - X.shape[1]
    lines = map(lambda v: ', '.join(process_column(v, pad)), X)

    if nr is None:
        nr = X.shape[1]
    else:
        nrq = int(math.ceil(nr / 4.0))
    if nc is None:
        nc = X.shape[0]

    fh.write('float {}[] = {}\n'.format('__' + name, '{'))
    fh.write('\t' + ',\n\t'.join(lines))
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', nr, nrq, nc, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))


def cformatV(fh, name, X):
    nrq = int(math.ceil(X.shape[0] / 4.0))
    pad = nrq * 4 - X.shape[0]
    lines = ', '.join(list(map(lambda f: small_hex(f), X)) + [small_hex(0.0)] * pad)
    fh.write('float {}[] = {}\n'.format( '__' + name, '{'))
    fh.write('\t' + lines)
    fh.write('};\n')
    fh.write('_Mat {} = {}\n\t.nr = {},\n\t.nrq = {},\n\t.nc = {},\n\t.stride = {},\n\t.data.f = {}\n{};\n'.format('_' + name, '{', X.shape[0], nrq, 1, nrq * 4, '__' + name, '}'))
    fh.write('const scrappie_matrix {} = &{};\n\n'.format(name, '_' + name))



with open(model_file, 'rb') as fh:
    network = pickle.load(fh, encoding='latin1')
assert network.version == 1, "Sloika model must be version 1.  Perhaps you need to run Sloika's model_upgrade.py"

sys.stdout.write("""#pragma once
#ifndef NANONET_RGR_MODEL_H
#define NANONET_RGR_MODEL_H
#include <assert.h>
#include "../util.h"
""")

""" Convolution layer
"""

filterW =  network.sublayers[0].W.get_value()
nfilter, _ , winlen = filterW.shape
cformatM(sys.stdout, 'conv_rgr_W', filterW.reshape(-1, 1), nr = winlen * 4 - 3, nc=nfilter)
cformatV(sys.stdout, 'conv_rgr_b', network.sublayers[0].b.get_value().reshape(-1))
sys.stdout.write("const int conv_rgr_stride = {};\n".format(network.sublayers[0].stride))
sys.stdout.write("""const size_t _conv_rgr_nfilter = {};
const size_t _conv_rgr_winlen = {};
""".format(nfilter, winlen))

"""  Backward GRU (first layer)
"""
gru1 = network.sublayers[1].sublayers[0]
cformatM(sys.stdout, 'gruB1_rgr_iW', gru1.iW.get_value())
cformatM(sys.stdout, 'gruB1_rgr_sW', gru1.sW.get_value())
cformatM(sys.stdout, 'gruB1_rgr_sW2', gru1.sW2.get_value())
cformatV(sys.stdout, 'gruB1_rgr_b', gru1.b.get_value().reshape(-1))

"""  Forward GRU (second layer)
"""
gru2 = network.sublayers[2]
cformatM(sys.stdout, 'gruF2_rgr_iW', gru2.iW.get_value())
cformatM(sys.stdout, 'gruF2_rgr_sW', gru2.sW.get_value())
cformatM(sys.stdout, 'gruF2_rgr_sW2', gru2.sW2.get_value())
cformatV(sys.stdout, 'gruF2_rgr_b', gru2.b.get_value().reshape(-1))



""" backward GRU(third layer)
"""
gru3 = network.sublayers[3].sublayers[0]
cformatM(sys.stdout, 'gruB3_rgr_iW', gru3.iW.get_value())
cformatM(sys.stdout, 'gruB3_rgr_sW', gru3.sW.get_value())
cformatM(sys.stdout, 'gruB3_rgr_sW2', gru3.sW2.get_value())
cformatV(sys.stdout, 'gruB3_rgr_b', gru3.b.get_value().reshape(-1))

""" Softmax layer
"""
nstate = network.sublayers[4].W.get_value().shape[0]
shuffle = np.append(np.arange(nstate - 1) + 1, 0)
cformatM(sys.stdout, 'FF_rgr_W', network.sublayers[4].W.get_value()[shuffle])
cformatV(sys.stdout, 'FF_rgr_b', network.sublayers[4].b.get_value()[shuffle])

sys.stdout.write('#endif /* NANONET_RGR_MODEL_H */')



import cPickle
import numpy as np
from sloika import features
from untangled import fast5

modelfile = 'model_final.pkl'
fn = 'reads/MINICOL228_20161012_FNFAB42578_MN17976_mux_scan_HG_52221_ch271_read66_strand.fast5'

with open(modelfile, 'r') as fh:
	network = cPickle.load(fh)

iWf = network.layers[1].layers[0].iW.get_value()
sWf = network.layers[1].layers[0].sW.get_value()
bf = network.layers[1].layers[0].b.get_value()
pf = network.layers[1].layers[0].p.get_value()

iWb = network.layers[1].layers[1].layer.iW.get_value()
sWb = network.layers[1].layers[1].layer.sW.get_value()
bb = network.layers[1].layers[1].layer.b.get_value()
pb = network.layers[1].layers[1].layer.p.get_value()

with fast5.Reader(fn) as f5:
    ev = f5.get_section_events("template", analysis='Segment_Linear')
    sn = f5.filename_short

ev = ev[50:]
print "Input"
print ev[:10]

features = features.from_events(ev, tag='')
print "Features"
print features[:10]

#  Windowing
w = 3
inMat = features
zeros = np.zeros((w // 2, features.shape[1]))
padMat = np.concatenate([zeros, inMat, zeros], axis=0)
tmp = np.concatenate([padMat[i : 1 + i - w] for i in xrange(w - 1)], axis=1)
feature3 = np.concatenate([tmp, padMat[w - 1 :]], axis=1)
print "Feature 3"
print feature3[:10]

#  LSTM forward
def gatefun(x):
	return 1.0 / (1.0 + np.exp(-x))
size = sWf.shape[1]
state = np.zeros(size)
lastout = np.zeros(size)
outF = np.zeros((len(feature3), size))
for i, f in enumerate(feature3):
        vW = np.tensordot(f, iWf, axes=(0, 1))
        outW = np.tensordot(lastout, sWf, axes=(0, 1))
        sumW = vW + outW  + bf
        sumW = sumW.reshape((size, 4))

        #  Forget gate activation
        forget = state * gatefun(sumW[:,2] + state * pf[1])
        #  Update state with input
        update = np.tanh(sumW[:,0]) * gatefun(sumW[:,1] + state * pf[0])
	state = forget + update
        #  Output gate activation
        lastout = np.tanh(state) * gatefun(sumW[:,3] + state * pf[2])
	outF[i] = lastout

print "LSTM forward"
print outF[:10,:8]

state = np.zeros(size)
lastout = np.zeros(size)
outB = np.zeros((len(feature3), size))
for i, f in enumerate(feature3[::-1]):
        vW = np.tensordot(f, iWb, axes=(0, 1))
        outW = np.tensordot(lastout, sWb, axes=(0, 1))
        sumW = vW + outW  + bb
        sumW = sumW.reshape((size, 4))

        #  Forget gate activation
        forget = state * gatefun(sumW[:,2] + state * pb[1])
        #  Update state with input
        update = np.tanh(sumW[:,0]) * gatefun(sumW[:,1] + state * pb[0])
	state = forget + update
        #  Output gate activation
        lastout = np.tanh(state) * gatefun(sumW[:,3] + state * pb[2])
	outB[i] = lastout
outB = outB[::-1]
print "LSTM backward"
print outB[:10,:8]

inFF = np.concatenate((outF, outB), axis=1)

W1 = network.layers[2].W.get_value()
b1 = network.layers[2].b.get_value()
outFF1 = np.tanh(np.tensordot(inFF, W1, axes=(1,1)) + b1)
print "FF1"
print outFF1[:10,:8]



from concurrent.futures import ThreadPoolExecutor
import functools
from io import StringIO
import os
import shutil
import subprocess
from timeit import default_timer as now
import unittest

import h5py
import numpy as np

import scrappy

_path_ = os.path.join(os.path.dirname(__file__), '..', '..', 'reads')
_test_reads_ = tuple(os.path.abspath(os.path.join(_path_, x)) for x in (
    'MINICOL228_20161012_FNFAB42578_MN17976_mux_scan_HG_52221_ch174_read172_strand.fast5',
    'MINICOL228_20161012_FNFAB42578_MN17976_mux_scan_HG_52221_ch271_read66_strand.fast5',
    'read_ch228_file118.fast5',
))
_test_fasta_ = {
    x:'{}.fa'.format(os.path.splitext(x)[0]) for x in _test_reads_
}

class TestScrappy(unittest.TestCase):

    def _parse_fasta(fh):
        # return name, seq pairs from a fh
        #   NB: assumes sequences are on a single line
        name, seq = None, None
        for i, line in enumerate(fh.readlines()):
            if i % 2 == 0:
                # remove ">" and take up to first whitespace
                name = os.path.basename(line[1:].split()[0])
            else:
                seq = line.rstrip()
                yield name, seq


    @classmethod
    def setUpClass(self):
        print("* TestScrappy")
        self.scrappie_exe = shutil.which('scrappie')
        if self.scrappie_exe is None:
            raise RuntimeError("Ensure scrappie is on PATH before running scrappy tests.")
        self.model = 'rgrgr_r94'
        self.expected_states = 1025
        self.expected_stride = 5

        # get scrappie basecalls for two test files
        cmd = ['scrappie', 'raw', '--model', self.model]
        cmd.extend(_test_reads_)
        scrappie_fasta = subprocess.check_output(cmd).decode()
        self.scrappie_seqs = dict()
        name, seq = None, None
        for name, seq in self._parse_fasta(StringIO(scrappie_fasta)):
            self.scrappie_seqs[name] = seq
            if len(self.scrappie_seqs) == len(_test_reads_):
                break

        # get signal data
        self.signals = dict(scrappy._raw_gen(_test_reads_))

        # just one signal and ref
        chosen_one = _test_reads_[0]
        self.one_signal = self.signals[os.path.basename(chosen_one)]
        with open(_test_fasta_[chosen_one], 'r') as fh:
            self.one_ref = next(self._parse_fasta(fh))[1]
        print("  finished setUp.")


    def test_000_same_as_scrappie(self):
        for fname, data in self.signals.items():
            seq, score, pos, _, _, _ = scrappy.basecall_raw(data, model=self.model)
            self.assertEqual(seq, self.scrappie_seqs[os.path.basename(fname)])


    def test_010_raw_table_type(self):
        # type should be correct
        for t in (np.float32, np.float64):
            rt = scrappy.RawTable(self.one_signal.astype(t))
            self.assertEqual(rt._data.dtype, scrappy.ftype, 'Raw table internal type.')


    def test_011_raw_table_methods(self):
        rt = scrappy.RawTable(self.one_signal)
        # check trim
        rt.trim(start=200)
        self.assertEqual(rt.start, 200, 'Trimming applied.')
        # check scale
        rt.scale()
        self.assertEqual(np.median(rt._data[rt.start:rt.end]), 0.0, 'Scaling shifts median to 0.0.')
        # .data(as_numpy=True) should own its data
        new_data = rt.data(as_numpy=True)
        self.assertIsInstance(new_data, np.ndarray)
        self.assertTrue(new_data.flags.owndata)


    def test_012_bad_trimming(self):
        rt = scrappy.RawTable(self.one_signal)
        rt.trim(start=200, end=len(rt._data) - 5)
        self.assertEqual(rt.start, 0, 'Empty gives start=0.')
        self.assertEqual(rt.end, 0, 'Empty gives end=0.')
        self.assertEqual(len(rt.data(as_numpy=True)), 0, 'Empty give len=0 array')


    def test_020_intermediates(self):
        rt = scrappy.RawTable(self.one_signal)
        self.assertIsInstance(rt._rt, scrappy.ffi.CData)
        rt.trim().scale()
        self.assertIsInstance(rt, scrappy.RawTable)
        self.assertIsInstance(rt._rt, scrappy.ffi.CData)
        post = scrappy.calc_post(rt, self.model, log=True)
        self.assertIsInstance(post, scrappy.ScrappyMatrix)

        # Check matrix is formed sanely
        sloika_post = scrappy._scrappie_to_numpy(post._data, sloika=True)
        self.assertIsInstance(sloika_post, np.ndarray)
        self.assertEqual(sloika_post.shape[1], self.expected_states)

        # check types, shouldn't leak cffi abstraction
        seq, score, pos = scrappy.decode_post(post, self.model)
        self.assertIsInstance(seq, str, 'sequence is str.')
        self.assertIsInstance(score, float, 'score is float.')
        self.assertIsInstance(pos, np.ndarray, 'pos is ndarray.')


    def test_030_threaded_call(self):
        # Just run this to check we don't die
        worker = functools.partial(scrappy.basecall_raw, model=self.model)
        with ThreadPoolExecutor(max_workers=1) as executor:
            results = executor.map(worker, self.signals.values())
            for res in results:
                pass


    def test_040_squiggle_map_r94(self):
        # Just check mapping runs without fail
        score, path = scrappy.map_signal_to_squiggle(self.one_signal, self.one_ref, model='squiggle_r94')
        self.assertIsInstance(score, float, 'score is float.')
        self.assertIsInstance(path, np.ndarray, 'path is ndarray.')
        self.assertEqual(len(self.one_signal), len(path), 'Length of path is length of signal.')

    def test_041_squiggle_map_r10(self):
        # Just check mapping runs without fail
        score, path = scrappy.map_signal_to_squiggle(self.one_signal, self.one_ref, model='squiggle_r10')
        self.assertIsInstance(score, float, 'score is float.')
        self.assertIsInstance(path, np.ndarray, 'path is ndarray.')

    def test_045_post_forward_mapping(self):
        rt = scrappy.RawTable(self.one_signal)
        rt.trim().scale()
        post = scrappy.calc_post(rt, self.model, log=True)

        t0 = now()
        score_band, _ = scrappy.map_post_to_sequence(
            post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
            viterbi=False, path=False, bands=100)
        t1 = now()
        score_no_band, _ = scrappy.map_post_to_sequence(
            post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
            viterbi=False, path=False, bands=None)
        t2 = now()
        self.assertIsInstance(score_no_band, float, 'score is float.')
        self.assertLess(t1 - t0, t2 - t0, 'banded mapping is faster.')

        with self.assertRaises(ValueError):
            # can't calculate path with Forward
            score_no_band = scrappy.map_post_to_sequence(
                post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
                viterbi=False, path=True, bands=None)


    def test_046_post_viterbi_mapping(self):
        rt = scrappy.RawTable(self.one_signal)
        rt.trim().scale()
        post = scrappy.calc_post(rt, self.model, log=True)

        t0 = now()
        score_band, _ = scrappy.map_post_to_sequence(
            post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
            viterbi=True, path=False, bands=100)
        t1 = now()
        score_no_band, _ = scrappy.map_post_to_sequence(
            post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
            viterbi=True, path=False, bands=None)
        t2 = now()
        self.assertIsInstance(score_no_band, float, 'score is float.')
        self.assertLess(t1 - t0, t2 - t0, 'banded mapping is faster.')

        score_band, path = scrappy.map_post_to_sequence(
            post, self.one_ref, stay_pen=0, skip_pen=0, local_pen=4.0,
            viterbi=True, path=True, bands=100)
        self.assertIsInstance(path, np.ndarray, 'path is ndarray.')


    def test_050_model_stride(self):
        stride = scrappy.get_model_stride(self.model)
        self.assertIsInstance(stride, int, 'stride is int.')
        self.assertEqual(stride, self.expected_stride, 'stride is as expected.')


    def test_051_model_stride_unknown(self):
        with self.assertRaises(ValueError):
            stride = scrappy.get_model_stride('garbage_model')


    def test_060_matrix_conversions(self):
        rt = scrappy.RawTable(self.one_signal)
        rt.trim().scale()
        original = scrappy.calc_post(rt, self.model, log=True)
        from_np = scrappy.ScrappyMatrix(
            original.data(as_numpy=True, sloika=False))

        # decoding both as a check on the C structure
        scores = [scrappy.decode_post(x)[1] for x in (original, from_np)]
        np.testing.assert_almost_equal(scores[0], scores[1], err_msg='Scores equal after round trip.')


    def test_065_matrix_view(self):
        rt = scrappy.RawTable(self.one_signal)
        rt.trim().scale()
        original = scrappy.calc_post(rt, self.model, log=True)

        all_view = original[:]
        self.assertSequenceEqual(original.shape, all_view.shape, 'All view is same shape.')
        small_view = original[100:200]
        self.assertSequenceEqual((100, original.shape[1]), small_view.shape, 'Slice has correct shape.')
        smaller_view = small_view[10:50]
        self.assertSequenceEqual((40, original.shape[1]), smaller_view.shape, 'Slice of slice has correct shape.')

        np_original = original.data(as_numpy=True, sloika=False)
        np_smaller = smaller_view.data(as_numpy=True, sloika=False)
        np.testing.assert_allclose(np_original[110:150,], np_smaller, err_msg='Slice contains correct data.')



